/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.factory.block;

import java.util.List;

import net.minecraft.block.Block;
import net.minecraft.block.material.Material;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.properties.PropertyBool;
import net.minecraft.block.properties.PropertyEnum;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.BlockRenderLayer;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.Axis;
import net.minecraft.util.EnumFacing.AxisDirection;
import net.minecraft.util.EnumHand;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;

import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import buildcraft.api.transport.pipe.ICustomPipeConnection;

import buildcraft.lib.block.BlockBCTile_Neptune;

import buildcraft.factory.tile.TileHeatExchangeEnd;
import buildcraft.factory.tile.TileHeatExchangeStart;

public class BlockHeatExchange extends BlockBCTile_Neptune implements ICustomPipeConnection {
    public enum Part {
        START,
        END,
        MIDDLE {
            @Override
            int getMeta(IBlockState state) {
                Axis axis = state.func_177229_b(PROP_AXIS);
                return axis == Axis.X ? 0 : 1;
            }

            @Override
            IBlockState getState(IBlockState defaultState, int meta) {
                return defaultState.func_177226_a(PROP_AXIS, (meta & 1) == 0 ? Axis.X : Axis.Z);
            }

            @Override
            IBlockState getPlacement(IBlockState state, EnumFacing playerFacing) {
                return state.func_177226_a(PROP_AXIS, playerFacing.func_176746_e().func_176740_k());
            }

            @Override
            public Axis getAxis(IBlockState state) {
                return state.func_177229_b(PROP_AXIS);
            }

            @Override
            void addProperties(List<IProperty<?>> properties) {
                properties.add(PROP_AXIS);
                properties.add(PROP_CONNECTED_NEG);
                properties.add(PROP_CONNECTED_POS);
            }

            @Override
            IBlockState getActualState(IBlockState state, IBlockAccess world, BlockPos pos) {
                Axis axis = getAxis(state);
                EnumFacing facePos = EnumFacing.func_181076_a(AxisDirection.POSITIVE, axis);
                state = state.func_177226_a(PROP_CONNECTED_POS, doesNeighbourConnect(world, pos, facePos));
                EnumFacing faceNeg = EnumFacing.func_181076_a(AxisDirection.NEGATIVE, axis);
                state = state.func_177226_a(PROP_CONNECTED_NEG, doesNeighbourConnect(world, pos, faceNeg));
                return state;
            }

            @Override
            boolean doesConnect(IBlockState state, EnumFacing from) {
                return from.func_176740_k() == state.func_177229_b(PROP_AXIS);
            }
        };

        int getMeta(IBlockState state) {
            EnumFacing face = state.func_177229_b(PROP_FACING);
            return face.func_176736_b();
        }

        IBlockState getState(IBlockState defaultState, int meta) {
            return defaultState.func_177226_a(PROP_FACING, EnumFacing.func_176731_b(meta & 3));
        }

        IBlockState getPlacement(IBlockState state, EnumFacing playerFacing) {
            EnumFacing face = playerFacing;
            if (this == END) face = face.func_176734_d();
            return state.func_177226_a(PROP_FACING, face.func_176746_e());
        }

        /** @return The axis That fluids flow through (horizontally) */
        public Axis getAxis(IBlockState state) {
            return state.func_177229_b(PROP_FACING).func_176740_k();
        }

        void addProperties(List<IProperty<?>> properties) {
            properties.add(PROP_FACING);
            properties.add(PROP_CONNECTED);
            properties.add(PROP_CONNECTED_Y);
        }

        IBlockState getActualState(IBlockState state, IBlockAccess world, BlockPos pos) {
            EnumFacing facing = state.func_177229_b(PROP_FACING);
            boolean connected = doesNeighbourConnect(world, pos, facing);
            state = state.func_177226_a(PROP_CONNECTED, connected);

            // EnumFacing yFace = this == Part.START ? EnumFacing.DOWN : EnumFacing.UP;
            // TODO: test the above connection (use the tile?)
            state = state.func_177226_a(PROP_CONNECTED_Y, Boolean.FALSE);
            return state;
        }

        boolean doesConnect(IBlockState state, EnumFacing from) {
            EnumFacing thisFacing = state.func_177229_b(PROP_FACING);
            return from == thisFacing;
        }

        static boolean doesNeighbourConnect(IBlockAccess world, BlockPos thisPos, EnumFacing dir) {
            IBlockState offset = world.func_180495_p(thisPos.func_177972_a(dir));
            Block block = offset.func_177230_c();
            if (block instanceof BlockHeatExchange) {
                Part part = ((BlockHeatExchange) block).part;
                return part.doesConnect(offset, dir.func_176734_d());
            }
            return false;
        }
    }

    public static final IProperty<Axis> PROP_AXIS = PropertyEnum.func_177706_a("axis", Axis.class, Axis.X, Axis.Z);
    public static final IProperty<Boolean> PROP_CONNECTED = PropertyBool.func_177716_a("connected");
    public static final IProperty<Boolean> PROP_CONNECTED_Y = PropertyBool.func_177716_a("connected_y");
    public static final IProperty<Boolean> PROP_CONNECTED_POS = PropertyBool.func_177716_a("connected_pos");
    public static final IProperty<Boolean> PROP_CONNECTED_NEG = PropertyBool.func_177716_a("connected_neg");

    private static Part currentInitPart = null;
    public final Part part;

    public BlockHeatExchange(Material material, String id, Part part) {
        // Java doesn't allow setting static variables here, so call a static method that does it inside super()
        super(material, setCurrentPart(id, part));
        currentInitPart = null;
        this.part = part;
    }

    private static String setCurrentPart(String id, Part part) {
        currentInitPart = part;
        return id;
    }

    @Override
    protected void addProperties(List<IProperty<?>> properties) {
        // This is called BEFORE part is set, so we have to use a static property instead
        currentInitPart.addProperties(properties);
    }

    @Override
    public int func_176201_c(IBlockState state) {
        return part.getMeta(state);
    }

    @Override
    public IBlockState func_176203_a(int meta) {
        return part.getState(func_176223_P(), meta);
    }

    @Override
    public IBlockState func_176221_a(IBlockState state, IBlockAccess world, BlockPos pos) {
        return part.getActualState(state, world, pos);
    }

    @Override
    public IBlockState getStateForPlacement(World world, BlockPos pos, EnumFacing facing, float hitX, float hitY, float hitZ, int meta, EntityLivingBase placer, EnumHand hand) {
        return part.getPlacement(func_176223_P(), placer.func_174811_aO());
    }

    @Override
    public TileEntity func_149915_a(World world, int meta) {
        switch (part) {
            case START:
                return new TileHeatExchangeStart();
            case MIDDLE:
                return null;
            case END:
                return new TileHeatExchangeEnd();
            default:
                throw new IllegalStateException("Unknown part " + part);
        }
    }

    @Override
    public boolean func_149662_c(IBlockState state) {
        return false;
    }

    @Override
    public boolean func_149686_d(IBlockState state) {
        return false;
    }

    @Override
    @SideOnly(Side.CLIENT)
    public BlockRenderLayer func_180664_k() {
        return BlockRenderLayer.CUTOUT;
    }

    @Override
    public float getExtension(World world, BlockPos pos, EnumFacing face, IBlockState state) {
        return 0;
    }
}
